from argparse import Namespace, ArgumentParser


def main_parser() -> Namespace:
    parser = ArgumentParser(prog='parity')

    commands = parser.add_subparsers(dest='cmd')

    exchanges = commands.add_parser('exchanges', help='list supported exchanges')
    exchanges.add_argument('-s', '--span', type=int, default=3)

    markets = commands.add_parser('markets', help='list supported markets')
    markets.add_argument('-s', '--span', type=int, default=3)
    markets.add_argument('--compact', action='store_true')

    ticker = commands.add_parser('ticker', help='fetch ticker from a given market')
    ticker.add_argument('exchange', type=str)
    ticker.add_argument('quote', type=str)
    ticker.add_argument('base', type=str)

    book = commands.add_parser('book', help='fetch book from a given market')
    book.add_argument('exchange', type=str)
    book.add_argument('quote', type=str)
    book.add_argument('base', type=str)

    return parser.parse_args()



from typing import List

import ccxt

from dotcoin.core.abstract import Exchange
from dotcoin.core.types import Market, Context, Ticker, Book
from dotcoin.sources.ccxt.utility import many_symbols_to_markets, parse_ticker, parse_book, market_to_symbol


class CCXTExchange(Exchange):
    source = "CCXT"

    def __init__(self, name: str, code: str):
        self.name = name
        self._ccxt_exchange: ccxt.Exchange = getattr(ccxt, code)()
        self._ccxt_exchange.load_markets()

    def markets(self) -> List[Market]:
        ccxt_symbols = self._ccxt_exchange.markets.keys()
        markets = many_symbols_to_markets(ccxt_symbols)
        return markets

    def has_market(self, market: Market) -> bool:
        ccxt_symbol = market_to_symbol(market)
        return ccxt_symbol in self._ccxt_exchange.markets

    def get_ticker(self, market: Market) -> Ticker:
        ccxt_symbol = market_to_symbol(market)
        raw_ticker = self._ccxt_exchange.fetch_ticker(ccxt_symbol)

        context = Context(market, self.name)
        ticker = parse_ticker(context, raw_ticker)
        return ticker

    def get_book(self, market: Market) -> Book:
        ccxt_symbol = market_to_symbol(market)
        raw_book = self._ccxt_exchange.fetch_l2_order_book(ccxt_symbol)

        context = Context(market, self.name)
        return parse_book(context, raw_book)



from collections import Counter
from itertools import groupby

from beautifultable import BeautifulTable

from dotcoin.core.types import Table


def content_padding(header, body, width):
    header = header + [''] * (width - len(header))
    body = [row + [''] * (width - len(row)) for row in body]
    return header, body


def content_chunks(items, width, parse=None):
    if parse:
        items = map(parse, items)
    items = list(items)
    return [items[i:i + width] for i in range(0, len(items), width)]


def content_fold(columns: dict):
    return [[key] + row for key, rows in columns.items() for row in rows]


class KeyTable:
    def __init__(self, header: list, items: list):
        self.header = header
        self.items = items
        self.key = None
        self.parser = None
        self.content_header = None
        self.content_body = None

    def set_key(self, key):
        self.key = key

    def set_parser(self, parser):
        self.parser = parser

    def build(self, span):
        counter = Counter(map(self.key, self.items))
        width = min(max(counter.values()), span)

        groups = groupby(self.items, self.key)
        span = {key: content_chunks(items, width, self.parser) for key, items in groups}

        body = content_fold(span)
        content_header, content_body = content_padding(self.header, body, width + 1)

        self.content_header = content_header
        self.content_body = content_body

    def get_table(self):
        header = self.content_header
        body = self.content_body
        return Table(header, body)


def print_table(table: Table) -> str:
    echo = BeautifulTable(detect_numerics=False)
    echo.columns.header = table.header
    for row in table.content:
        echo.rows.append(row)

    echo.columns.alignment = BeautifulTable.ALIGN_LEFT
    echo.set_style(BeautifulTable.STYLE_BOX)
    echo.columns.header.separator = '═'
    echo.border.header_left = '╞'
    echo.border.header_right = '╡'
    echo.columns.header.junction = '╪'
    return str(echo)



from typing import List
from dataclasses import dataclass


@dataclass(frozen=True)
class Market:
    quote: str
    base: str


@dataclass
class Context:
    market: Market
    exchange: str


@dataclass
class Ticker:
    context: Context
    last: float
    bid: float
    ask: float
    volume: float
    var: float


@dataclass
class Order:
    quote: float
    base: float
    price: float


@dataclass
class Book:
    context: Context
    bids: List[Order]
    asks: List[Order]


@dataclass
class Table:
    header: List[str]
    content: List[List[str]]



from typing import List

from dotcoin.context.fetch import fetch_ticker, fetch_book
from dotcoin.context.source import all_exchanges
from dotcoin.core.types import Market, Ticker, Book


def all_markets() -> List[Market]:
    markets = set()
    for exchange in all_exchanges():
        markets.update(exchange.markets())
    return list(markets)


def get_ticker(exchange: str, quote: str, base: str) -> Ticker:
    market = Market(quote.upper(), base.upper())
    return fetch_ticker(exchange.upper(), market)


def get_book(exchange: str, quote: str, base: str) -> Book:
    market = Market(quote.upper(), base.upper())
    return fetch_book(exchange.upper(), market)



from dotcoin.queries.index import index_markets, index_exchanges
from dotcoin.queries.markets import get_ticker, get_book

from dotcoin.tables.index import table_markets, table_exchanges
from dotcoin.tables.ticker import table_ticker
from dotcoin.tables.book import table_book
from dotcoin.tables.utility import print_table


def print_exchanges(span: int):
    items = index_exchanges()
    table = table_exchanges(items, span=span)
    output = print_table(table)
    print(output)


def print_markets(span: int, compact: bool):
    items = index_markets()
    table = table_markets(items, span=span, compact=compact)
    output = print_table(table)
    print(output)


def print_ticker(exchange: str, quote: str, base: str):
    item = get_ticker(exchange, quote, base)
    table = table_ticker(item)
    output = print_table(table)
    print(output)


def print_book(exchange: str, quote: str, base: str):
    item = get_book(exchange, quote, base)
    table = table_book(item)
    output = print_table(table)
    print(output)



from dotcoin.core.types import Market, Order, Book, Table

_HEADER = "{quote}|{base}|PRICE|PRICE|{base}|{quote}"
_BID_ORDER = "{quote:>8f}|{base:>8f}|{price:>.8f}"
_ASK_ORDER = "{price:>8f}|{base:>8f}|{quote:>.8f}"


def _format_header(market: Market):
    header = _HEADER.format(**market.__dict__)
    return header.split("|")


def _format_order(bid: Order, ask: Order):
    bid_fmt = _BID_ORDER.format(**bid.__dict__)
    ask_fmt = _ASK_ORDER.format(**ask.__dict__)
    order_row = [*bid_fmt.split("|"), *ask_fmt.split("|")]
    return order_row


def _format_content(book: Book):
    zipped = zip(book.bids, book.asks)
    body = [_format_order(bid, ask) for bid, ask in zipped]
    return body


def table_book(book: Book) -> Table:
    header = _format_header(book.context.market)
    content = _format_content(book)
    return Table(header, content)



from abc import ABC, abstractmethod
from typing import List

from dotcoin.core.types import Market, Ticker, Book


class Exchange(ABC):
    name: str = None
    source: str = None

    @abstractmethod
    def markets(self) -> List[Market]:
        raise NotImplementedError

    @abstractmethod
    def has_market(self, market: Market) -> bool:
        raise NotImplementedError

    @abstractmethod
    def get_ticker(self, market: Market) -> Ticker:
        raise NotImplementedError

    @abstractmethod
    def get_book(self, market: Market) -> Book:
        raise NotImplementedError


class Source(ABC):
    name: str = None

    @staticmethod
    @abstractmethod
    def exchanges() -> List[str]:
        raise NotImplementedError

    @staticmethod
    @abstractmethod
    def has_exchange(exchange: str) -> bool:
        raise NotImplementedError

    @staticmethod
    @abstractmethod
    def get_exchange(exchange: str) -> Exchange:
        raise NotImplementedError



from typing import Iterator, List

from dotcoin.core.abstract import Exchange
from dotcoin.context.source import get_exchange


def exchanges_to_names(exchanges: Iterator[Exchange]) -> List[str]:
    return [getattr(exchange, 'name') for exchange in exchanges]


def names_to_exchanges(names: Iterator[str]) -> List[Exchange]:
    return [get_exchange(name.upper()) for name in names]



from typing import Dict, List

from dotcoin.core.abstract import Exchange, Source
from dotcoin.core.error import InvalidSourceError
from dotcoin.sources import all_sources

_exchange_source_map_ = {}


def _source_map() -> Dict[str, Source]:
    if not _exchange_source_map_:
        for source in all_sources():
            for name in source.exchanges():
                if name not in _exchange_source_map_:
                    _exchange_source_map_[name] = source
    return _exchange_source_map_


def all_exchanges() -> List[Exchange]:
    return [source.get_exchange(name) for name, source in _source_map().items()]


def has_exchange(exchange: str) -> bool:
    return exchange in _source_map()


def get_exchange(exchange: str) -> Exchange:
    if has_exchange(exchange):
        source: Source = _source_map().get(exchange)
        return source.get_exchange(exchange)
    raise InvalidSourceError(f'"{exchange}" is not a supported exchange.')



from typing import List, Iterator
from collections import Counter

from dotcoin.core.types import Market
from dotcoin.core.abstract import Exchange


def match_markets(exchanges: List[Exchange]) -> Iterator[Market]:
    counter = Counter()
    for exchange in exchanges:
        markets = exchange.markets()
        counter.update(markets)

    quota = len(exchanges)
    return filter(lambda match: counter[match] is quota, counter)


def match_exchanges(market: Market, exchanges: List[Exchange]) -> Iterator[Exchange]:
    return filter(lambda exchange: exchange.has_market(market), exchanges)



from typing import List

from dotcoin.core.types import Market
from dotcoin.context.source import all_exchanges, get_exchange
from dotcoin.queries.markets import all_markets
from dotcoin.queries.utility import exchanges_to_names


def index_exchanges() -> List[str]:
    return exchanges_to_names(all_exchanges())


def index_markets() -> List[Market]:
    markets = all_markets()
    return sorted(markets, key=str)


def index_markets_from(exchange: str) -> List[Market]:
    _exchange = get_exchange(exchange)
    markets = _exchange.markets()
    return sorted(markets, key=str)



from typing import List

from dotcoin.core.types import Market
from dotcoin.context.match import match_markets, match_exchanges
from dotcoin.context.source import all_exchanges
from dotcoin.queries.utility import exchanges_to_names, names_to_exchanges


def common_exchanges(quote: str, base: str) -> List[str]:
    market = Market(quote.upper(), base.upper())
    matches = match_exchanges(market, all_exchanges())
    names = exchanges_to_names(matches)
    return sorted(names)


def common_markets(exchanges: List[str]) -> List[Market]:
    _exchanges = names_to_exchanges(exchanges)
    matches = match_markets(_exchanges)
    return sorted(matches, key=str)



from typing import List

from dotcoin.core.types import Market, Context, Ticker, Order, Book


def market_to_symbol(market: Market) -> str:
    return f"{market.base}/{market.quote}"


def symbol_to_market(symbol: str) -> Market:
    base, quote = symbol.split("/")
    return Market(quote, base)


def many_symbols_to_markets(symbols: list) -> List[Market]:
    return list(map(symbol_to_market, symbols))


TICKER_KEYMAP = {'bid': 'bid', 'ask': 'ask', 'last': 'last', 'volume': 'baseVolume', 'var': 'percentage'}


def parse_ticker(context: Context, ticker: dict) -> Ticker:
    str_dict = {next_key: ticker[old_key] for next_key, old_key in TICKER_KEYMAP.items()}
    num_dict = {key: float(value) for key, value in str_dict.items()}
    return Ticker(context, **num_dict)


def parse_order(item: list) -> Order:
    price, base = item
    return Order(price * base, base, price)


def parse_book(context: Context, book: dict) -> Book:
    bids = list(map(parse_order, book['bids']))
    asks = list(map(parse_order, book['asks']))
    return Book(context, bids, asks)



from typing import List
from pathlib import Path
from importlib import import_module

from dotcoin.core.abstract import Source

SOURCE_PATH = Path(__path__[0])


def find_source(name: str) -> Source.__class__:
    module = import_module(f"{__name__}.{name}.source")
    for key in dir(module):
        if key.lower() == f"{name}source":
            return getattr(module, key)


def all_sources() -> List[Source.__class__]:
    source_list = []
    for item in SOURCE_PATH.glob("*/"):
        if not item.name.startswith("_"):
            source = find_source(item.name)
            if issubclass(source, Source):
                source_list.append(source)
    return source_list



from dotcoin.core.types import Table, Market, Ticker

_HEADER = "Last {base}/{quote}|Volume {base}|Var.|Bid {base}/{quote}|Ask {quote}/{base}"
_CONTENT = "{last}|{volume}|{var:.2f}%|{bid}|{ask}"


def _format_header(market: Market):
    header = _HEADER.format(**market.__dict__)
    return header.split("|")


def _format_content(ticker: Ticker):
    body = _CONTENT.format(**ticker.__dict__)
    return body.split("|")


def table_ticker(ticker: Ticker) -> Table:
    header = _format_header(ticker.context.market)
    content = _format_content(ticker)
    return Table(header, [content])



from dotcoin.console.parser import main_parser
from dotcoin.console.actions import print_exchanges, print_markets, print_ticker, print_book


def main():
    args = main_parser()
    if args.cmd == 'exchanges':
        print_exchanges(args.span)
    elif args.cmd == 'markets':
        print_markets(args.span, args.compact)
    elif args.cmd == 'ticker':
        print_ticker(args.exchange, args.quote, args.base)
    elif args.cmd == 'book':
        print_book(args.exchange, args.quote, args.base)


__all__ = {'main'}



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['dotcoin',
 'dotcoin.console',
 'dotcoin.context',
 'dotcoin.core',
 'dotcoin.queries',
 'dotcoin.sources',
 'dotcoin.sources.ccxt',
 'dotcoin.tables']

package_data = \
{'': ['*']}

install_requires = \
['beautifultable>=1.0.1,<2.0.0', 'ccxt>=1.47.3,<2.0.0', 'six>=1.15.0,<2.0.0']

entry_points = \
{'console_scripts': ['dotcoin = dotcoin.__main__:main']}

setup_kwargs = {
    'name': 'dotcoin',
    'version': '0.1.0',
    'description': '',
    'long_description': '# DotCoin\n\n## Introduction\n\nDotCoin is an abstract interface build around ccxt to access information on cryptocurrency markets from a variety of\nexchanges/platforms, while being accessible to non-coders or coders who don\'t want to write a program just to check some\nbasic info.\n\nAs of this moment, DotCoin provides a simple cli that can look into the basic stats of some markets and checking their\ncurrent order-book. Please bear in mind that DotCoin is currently just a personal project that just started its\ndevelopment status and needs a lot of time and work to grow.\n\n## Installation\n\nA package with the name "dotcoin" is available in pypi.\n\n````commandline\npip install --user dotcoin \n````\n\nThis package has been tested mainly with **Python 3.7.9** on **Windows 10**, but it should work on all platforms since\nit only makes use of cross-platform libraries. You may need to prefix this command with *"pip3"*, *"python3 -m pip"* or\n*"python -m pip"* instead when running on other platforms or different Python setups.\n\n## General Usage\n\n````commandline\ndotcoin -h\n````\n\n````text\nusage: parity [-h] {exchanges,markets,ticker,book} ...\n\npositional arguments:\n  {exchanges,markets,ticker,book}\n    exchanges           list supported exchanges\n    markets             list supported markets\n    ticker              fetch ticker from a given market\n    book                fetch book from a given market\n\noptional arguments:\n  -h, --help            show this help message and exit\n````\n\n## Limited support\n\nAs of now, only these exchanges are supported.\n\n* BINANCE\n* BITTREX\n\nAlthough DotCoin is build on top of ccxt, which supports a considerable number of platforms, as interface which targets\nthe end-user, I consider it\'s important to provide a stable usability instead of unreliable feature set. This is why\nthere will be limited support for most exchanges in the meantime, and more will be added once certain level of\nreliability is achieved.\n\n## Thanks\n\nI greatly thank the ccxt team for their amazing work on the ccxt library, it\'s only thanks to the project they\'ve built\nthat this app is even possible. You can check their work out at their [GitHub](https://github.com/ccxt/ccxt).\n',
    'author': 'Orlando Ospino Sánchez',
    'author_email': 'oroschz@outlook.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/oroschz/dotcoin',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



from typing import List

from dotcoin.core.abstract import Exchange, Source
from dotcoin.sources.ccxt.exchange import CCXTExchange

SUPPORTED = {
    "BINANCE": "binance",
    "BITTREX": "bittrex"
}


class CCXTSource(Source):
    name = "ccxt"

    @staticmethod
    def exchanges() -> List[str]:
        exchanges = list(SUPPORTED.keys())
        return exchanges

    @staticmethod
    def has_exchange(exchange: str) -> bool:
        return exchange in SUPPORTED

    @staticmethod
    def get_exchange(exchange: str) -> Exchange:
        codename = SUPPORTED[exchange]
        return CCXTExchange(exchange, codename)



from dotcoin.core.error import InvalidSourceError, InvalidContextError
from dotcoin.core.types import Market, Ticker, Book
from dotcoin.context.source import has_exchange, get_exchange


def fetch_ticker(exchange: str, market: Market) -> Ticker:
    if has_exchange(exchange):
        _exchange = get_exchange(exchange)
        if _exchange.has_market(market):
            return _exchange.get_ticker(market)
        raise InvalidContextError(f'"{market}" is not supported for this exchange.')
    raise InvalidSourceError(f'"{exchange}" is not a supported exchange.')


def fetch_book(exchange: str, market: Market) -> Book:
    if has_exchange(exchange):
        _exchange = get_exchange(exchange)
        if _exchange.has_market(market):
            return _exchange.get_book(market)
        raise InvalidContextError(f'"{market}" is not supported for this exchange.')
    raise InvalidSourceError(f'"{exchange}" is not a supported exchange.')



from typing import List

from dotcoin.core.types import Table, Market
from dotcoin.tables.utility import KeyTable


def table_exchanges(exchanges: List[str], span: int = 3) -> Table:
    headers = ['Init.', 'Exchanges']

    content = KeyTable(headers, exchanges)
    content.set_key(lambda word: word[0])
    content.build(span)

    return content.get_table()


def table_markets(markets: List[Market], span: int = 3, compact: bool = False) -> Table:
    headers = ['Quote', 'Base']

    content = KeyTable(headers, markets)
    content.set_key(lambda market: market.quote)
    if compact:
        content.set_parser(lambda market: f"{market.base}")
    else:
        content.set_parser(lambda market: f"{market.base}/{market.quote}")
    content.build(span)

    return content.get_table()
